//  panel.cxx - default, 2D single-engine prop instrument panel
//
//  Written by David Megginson, started January 2000.
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of the
//  License, or (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  $Id: panel.cxx,v 1.21 2000/09/13 19:51:07 curt Exp $

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_WINDOWS_H          
#  include <windows.h>
#endif

#include <string.h>

#include <plib/ssg.h>
#include <plib/fnt.h>

#include <simgear/debug/logstream.hxx>
#include <simgear/misc/fgpath.hxx>
#include <Main/options.hxx>
#include <Main/views.hxx>
#include <Objects/texload.h>

#include "hud.hxx"
#include "panel.hxx"



////////////////////////////////////////////////////////////////////////
// Implementation of FGTextureManager.
////////////////////////////////////////////////////////////////////////

map<string,ssgTexture *> FGTextureManager::_textureMap;

ssgTexture *
FGTextureManager::createTexture (const string &relativePath)
{
  ssgTexture * texture = _textureMap[relativePath];
  if (texture == 0) {
    cerr << "Texture " << relativePath << " does not yet exist" << endl;
    FGPath tpath(current_options.get_fg_root());
    tpath.append(relativePath);
    texture = new ssgTexture((char *)tpath.c_str(), false, false);
    _textureMap[relativePath] = texture;
    if (_textureMap[relativePath] == 0) 
      cerr << "Texture *still* doesn't exist" << endl;
    cerr << "Created texture " << relativePath
	 << " handle=" << texture->getHandle() << endl;
  }

  return texture;
}



////////////////////////////////////////////////////////////////////////
// Implementation of FGPanel.
////////////////////////////////////////////////////////////////////////

FGPanel * current_panel = NULL;

FGPanel::FGPanel (int x, int y, int w, int h)
  : _mouseDown(false),
    _mouseInstrument(0),
    _x(x), _y(y), _w(w), _h(h)
{
  setVisibility(current_options.get_panel_status());
  _panel_h = (int)(h * 0.5768 + 1);
}

FGPanel::~FGPanel ()
{
  for (instrument_list_type::iterator it = _instruments.begin();
       it != _instruments.end();
       it++) {
    delete *it;
    *it = 0;
  }
}

void
FGPanel::addInstrument (FGPanelInstrument * instrument)
{
  _instruments.push_back(instrument);
}

void
FGPanel::update () const
{
				// Do nothing if the panel isn't visible.
  if (!_visibility)
    return;

				// If the mouse is down, do something
  if (_mouseDown) {
    _mouseDelay--;
    if (_mouseDelay < 0) {
      _mouseInstrument->doMouseAction(_mouseButton, _mouseX, _mouseY);
      _mouseDelay = 2;
    }
  }

				// Now, draw the panel
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  gluOrtho2D(_x, _x + _w, _y, _y + _h);

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();

				// Draw the background
  glEnable(GL_TEXTURE_2D);
  glDisable(GL_LIGHTING);
  glEnable(GL_BLEND);
  glEnable(GL_ALPHA_TEST);
  glEnable(GL_COLOR_MATERIAL);
  // glColor4f(1.0, 1.0, 1.0, 1.0);
  if ( cur_light_params.sun_angle * RAD_TO_DEG < 95.0 ) {
      glColor4fv( cur_light_params.scene_diffuse );
  } else {
      glColor4f(0.7, 0.2, 0.2, 1.0);
  }
  glBindTexture(GL_TEXTURE_2D, _bg->getHandle());
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glBegin(GL_POLYGON);
  glTexCoord2f(0.0, 0.0); glVertex3f(_x, _y, 0);
  glTexCoord2f(10.0, 0.0); glVertex3f(_x + _w, _y, 0);
  glTexCoord2f(10.0, 5.0); glVertex3f(_x + _w, _y + _panel_h, 0);
  glTexCoord2f(0.0, 5.0); glVertex3f(_x, _y + _panel_h, 0);
  glEnd();

				// Draw the instruments.
  instrument_list_type::const_iterator current = _instruments.begin();
  instrument_list_type::const_iterator end = _instruments.end();

  for ( ; current != end; current++) {
    FGPanelInstrument * instr = *current;
    glLoadIdentity();
    glTranslated(instr->getXPos(), instr->getYPos(), 0);
    instr->draw();
  }

  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
  ssgForceBasicState();
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
}

void
FGPanel::setVisibility (bool visibility)
{
  _visibility = visibility;
}

bool
FGPanel::getVisibility () const
{
  return _visibility;
}

void
FGPanel::setBackground (ssgTexture * texture)
{
  _bg = texture;
}

bool
FGPanel::doMouseAction (int button, int updown, int x, int y)
{
				// Note a released button and return
  // cerr << "Doing mouse action\n";
  if (updown == 1) {
    _mouseDown = false;
    _mouseInstrument = 0;
    return true;
  }

  x = (int)(((float)x / current_view.get_winWidth()) * _w);
  y = (int)(_h - (((float)y / current_view.get_winHeight()) * _h));

  for (int i = 0; i < (int)_instruments.size(); i++) {
    FGPanelInstrument *inst = _instruments[i];
    int ix = inst->getXPos();
    int iy = inst->getYPos();
    int iw = inst->getWidth() / 2;
    int ih = inst->getHeight() / 2;
    if (x >= ix - iw && x < ix + iw && y >= iy - ih && y < iy + ih) {
      _mouseDown = true;
      _mouseDelay = 20;
      _mouseInstrument = inst;
      _mouseButton = button;
      _mouseX = x - ix;
      _mouseY = y - iy;
				// Always do the action once.
      _mouseInstrument->doMouseAction(_mouseButton, _mouseX, _mouseY);
      return true;
    }
  }
  return false;
}



////////////////////////////////////////////////////////////////////////.
// Implementation of FGPanelAction.
////////////////////////////////////////////////////////////////////////

FGPanelAction::FGPanelAction ()
{
}

FGPanelAction::FGPanelAction (int button, int x, int y, int w, int h)
  : _button(button), _x(x), _y(y), _w(w), _h(h)
{
}

FGPanelAction::~FGPanelAction ()
{
}



////////////////////////////////////////////////////////////////////////
// Implementation of FGAdjustAction.
////////////////////////////////////////////////////////////////////////

FGAdjustAction::FGAdjustAction (int button, int x, int y, int w, int h,
				SGValue * value, float increment, 
				float min, float max, bool wrap)
  : FGPanelAction(button, x, y, w, h),
    _value(value), _increment(increment), _min(min), _max(max), _wrap(wrap)
{
}

FGAdjustAction::~FGAdjustAction ()
{
}

void
FGAdjustAction::doAction ()
{
  float val = _value->getFloatValue();
  val += _increment;
  if (val < _min) {
    val = (_wrap ? _max : _min);
  } else if (val > _max) {
    val = (_wrap ? _min : _max);
  }
  _value->setDoubleValue(val);
}



////////////////////////////////////////////////////////////////////////
// Implementation of FGSwapAction.
////////////////////////////////////////////////////////////////////////

FGSwapAction::FGSwapAction (int button, int x, int y, int w, int h,
			    SGValue * value1, SGValue * value2)
  : FGPanelAction(button, x, y, w, h), _value1(value1), _value2(value2)
{
}

FGSwapAction::~FGSwapAction ()
{
}

void
FGSwapAction::doAction ()
{
  float val = _value1->getFloatValue();
  _value1->setDoubleValue(_value2->getFloatValue());
  _value2->setDoubleValue(val);
}



////////////////////////////////////////////////////////////////////////
// Implementation of FGToggleAction.
////////////////////////////////////////////////////////////////////////

FGToggleAction::FGToggleAction (int button, int x, int y, int w, int h,
				SGValue * value)
  : FGPanelAction(button, x, y, w, h), _value(value)
{
}

FGToggleAction::~FGToggleAction ()
{
}

void
FGToggleAction::doAction ()
{
  _value->setBoolValue(!(_value->getBoolValue()));
}



////////////////////////////////////////////////////////////////////////
// Implementation of FGPanelTransformation.
////////////////////////////////////////////////////////////////////////

FGPanelTransformation::FGPanelTransformation ()
{
}

FGPanelTransformation::FGPanelTransformation (Type _type,
					      const SGValue * _value,
					      float _min, float _max,
					      float _factor, float _offset)
  : type(_type), value(_value), min(_min), max(_max),
    factor(_factor), offset(_offset)
{
}

FGPanelTransformation::~FGPanelTransformation ()
{
}



////////////////////////////////////////////////////////////////////////
// Implementation of FGPanelInstrument.
////////////////////////////////////////////////////////////////////////


FGPanelInstrument::FGPanelInstrument ()
{
  setPosition(0, 0);
  setSize(0, 0);
}

FGPanelInstrument::FGPanelInstrument (int x, int y, int w, int h)
{
  setPosition(x, y);
  setSize(w, h);
}

FGPanelInstrument::~FGPanelInstrument ()
{
  for (action_list_type::iterator it = _actions.begin();
       it != _actions.end();
       it++) {
    delete *it;
    *it = 0;
  }
}

void
FGPanelInstrument::setPosition (int x, int y)
{
  _x = x;
  _y = y;
}

void
FGPanelInstrument::setSize (int w, int h)
{
  _w = w;
  _h = h;
}

int
FGPanelInstrument::getXPos () const
{
  return _x;
}

int
FGPanelInstrument::getYPos () const
{
  return _y;
}

int
FGPanelInstrument::getWidth () const
{
  return _w;
}

int
FGPanelInstrument::getHeight () const
{
  return _h;
}

void
FGPanelInstrument::addAction (FGPanelAction * action)
{
  _actions.push_back(action);
}

				// Coordinates relative to centre.
bool
FGPanelInstrument::doMouseAction (int button, int x, int y)
{
  action_list_type::iterator it = _actions.begin();
  action_list_type::iterator last = _actions.end();
  for ( ; it != last; it++) {
    if ((*it)->inArea(button, x, y)) {
      (*it)->doAction();
      return true;
    }
  }
  return false;
}



////////////////////////////////////////////////////////////////////////
// Implementation of FGLayeredInstrument.
////////////////////////////////////////////////////////////////////////

FGLayeredInstrument::FGLayeredInstrument (int x, int y, int w, int h)
  : FGPanelInstrument(x, y, w, h)
{
}

FGLayeredInstrument::~FGLayeredInstrument ()
{
  for (layer_list::iterator it = _layers.begin(); it != _layers.end(); it++) {
    delete *it;
    *it = 0;
  }
}

void
FGLayeredInstrument::draw ()
{
  for (int i = 0; i < (int)_layers.size(); i++) {
    glPushMatrix();
    glTranslatef(0.0, 0.0, (i / 100.0) + 0.1);
    _layers[i]->draw();
    glPopMatrix();
  }
}

int
FGLayeredInstrument::addLayer (FGInstrumentLayer *layer)
{
  int n = _layers.size();
  if (layer->getWidth() == -1) {
    layer->setWidth(getWidth());
  }
  if (layer->getHeight() == -1) {
    layer->setHeight(getHeight());
  }
  _layers.push_back(layer);
  return n;
}

int
FGLayeredInstrument::addLayer (CroppedTexture &texture,
			       int w, int h)
{
  return addLayer(new FGTexturedLayer(texture, w, h));
}

void
FGLayeredInstrument::addTransformation (FGPanelTransformation * transformation)
{
  int layer = _layers.size() - 1;
  _layers[layer]->addTransformation(transformation);
}



////////////////////////////////////////////////////////////////////////
// Implementation of FGInstrumentLayer.
////////////////////////////////////////////////////////////////////////

FGInstrumentLayer::FGInstrumentLayer (int w, int h)
  : _w(w),
    _h(h)
{
}

FGInstrumentLayer::~FGInstrumentLayer ()
{
  for (transformation_list::iterator it = _transformations.begin();
       it != _transformations.end();
       it++) {
    delete *it;
    *it = 0;
  }
}

void
FGInstrumentLayer::transform () const
{
  transformation_list::const_iterator it = _transformations.begin();
  transformation_list::const_iterator last = _transformations.end();
  while (it != last) {
    FGPanelTransformation *t = *it;
    float val = (t->value == 0 ? 0.0 : t->value->getFloatValue());
    if (val < t->min) {
      val = t->min;
    } else if (val > t->max) {
      val = t->max;
    }
    val = val * t->factor + t->offset;

    switch (t->type) {
    case FGPanelTransformation::XSHIFT:
      glTranslatef(val, 0.0, 0.0);
      break;
    case FGPanelTransformation::YSHIFT:
      glTranslatef(0.0, val, 0.0);
      break;
    case FGPanelTransformation::ROTATION:
      glRotatef(-val, 0.0, 0.0, 1.0);
      break;
    }
    it++;
  }
}

void
FGInstrumentLayer::addTransformation (FGPanelTransformation * transformation)
{
  _transformations.push_back(transformation);
}



////////////////////////////////////////////////////////////////////////
// Implementation of FGTexturedLayer.
////////////////////////////////////////////////////////////////////////


FGTexturedLayer::FGTexturedLayer (const CroppedTexture &texture, int w, int h)
  : FGInstrumentLayer(w, h)
{
  setTexture(texture);
}


FGTexturedLayer::~FGTexturedLayer ()
{
}


void
FGTexturedLayer::draw ()
{
  int w2 = _w / 2;
  int h2 = _h / 2;

  transform();
  glBindTexture(GL_TEXTURE_2D, _texture.texture->getHandle());
  glBegin(GL_POLYGON);
  if ( cur_light_params.sun_angle * RAD_TO_DEG < 95.0 ) {
      glColor4fv( cur_light_params.scene_diffuse );
  } else {
      glColor4f(0.7, 0.2, 0.2, 1.0);
  }
  glTexCoord2f(_texture.minX, _texture.minY); glVertex2f(-w2, -h2);
  glTexCoord2f(_texture.maxX, _texture.minY); glVertex2f(w2, -h2);
  glTexCoord2f(_texture.maxX, _texture.maxY); glVertex2f(w2, h2);
  glTexCoord2f(_texture.minX, _texture.maxY); glVertex2f(-w2, h2);
  glEnd();
}



////////////////////////////////////////////////////////////////////////
// Implementation of FGWindowLayer.
////////////////////////////////////////////////////////////////////////

FGWindowLayer::FGWindowLayer (int w, int h)
  : FGTexturedLayer (w, h)
{
}

FGWindowLayer::FGWindowLayer (const CroppedTexture &texture, int w, int h)
  : FGTexturedLayer(texture, w, h)
{
}

FGWindowLayer::~FGWindowLayer ()
{
}

void
FGWindowLayer::draw ()
{
  // doesn't do anything yet
  FGTexturedLayer::draw();
}



////////////////////////////////////////////////////////////////////////
// Implementation of FGTextLayer.
////////////////////////////////////////////////////////////////////////

FGTextLayer::FGTextLayer (int w, int h, Chunk * chunk1, Chunk * chunk2,
			  Chunk * chunk3)
  : FGInstrumentLayer(w, h), _pointSize(14.0)
{
  _color[0] = _color[1] = _color[2] = 0.0;
  _color[3] = 1.0;
  if (chunk1)
    addChunk(chunk1);
  if (chunk2)
    addChunk(chunk2);
  if (chunk3)
    addChunk(chunk3);
}

FGTextLayer::~FGTextLayer ()
{
  chunk_list::iterator it = _chunks.begin();
  chunk_list::iterator last = _chunks.end();
  for ( ; it != last; it++) {
    delete *it;
  }
}

void
FGTextLayer::draw ()
{
  glPushMatrix();
  glColor4fv(_color);
  transform();
  _renderer.setFont(guiFntHandle);
  _renderer.setPointSize(_pointSize);
  _renderer.begin();
  _renderer.start3f(0, 0, 0);

				// Render each of the chunks.
  chunk_list::const_iterator it = _chunks.begin();
  chunk_list::const_iterator last = _chunks.end();
  for ( ; it != last; it++) {
    _renderer.puts((char *)((*it)->getValue()));
  }

  _renderer.end();
  glColor4f(1.0, 1.0, 1.0, 1.0);	// FIXME
  glPopMatrix();
}

void
FGTextLayer::addChunk (FGTextLayer::Chunk * chunk)
{
  _chunks.push_back(chunk);
}

void
FGTextLayer::setColor (float r, float g, float b)
{
  _color[0] = r;
  _color[1] = g;
  _color[2] = b;
  _color[3] = 1.0;
}

void
FGTextLayer::setPointSize (float size)
{
  _pointSize = size;
}

void
FGTextLayer::setFont(fntFont * font)
{
  _renderer.setFont(font);
}



////////////////////////////////////////////////////////////////////////
// Implementation of FGTextLayer::Chunk.
////////////////////////////////////////////////////////////////////////

FGTextLayer::Chunk::Chunk (const string &text, const string &fmt)
  : _type(FGTextLayer::TEXT), _fmt(fmt)
{
  _text = text;
  if (_fmt == "") 
    _fmt = "%s";
}

FGTextLayer::Chunk::Chunk (ChunkType type, const SGValue * value,
			   const string &fmt, float mult)
  : _type(type), _fmt(fmt), _mult(mult)
{
  if (_fmt == "") {
    if (type == TEXT_VALUE)
      _fmt = "%s";
    else
      _fmt = "%.2f";
  }
  _value = value;
}

const char *
FGTextLayer::Chunk::getValue () const
{
  switch (_type) {
  case TEXT:
    sprintf(_buf, _fmt.c_str(), _text.c_str());
    return _buf;
  case TEXT_VALUE:
    sprintf(_buf, _fmt.c_str(), _value->getStringValue().c_str());
    break;
  case DOUBLE_VALUE:
    sprintf(_buf, _fmt.c_str(), _value->getFloatValue() * _mult);
    break;
  }
  return _buf;
}



////////////////////////////////////////////////////////////////////////
// Implementation of FGSwitchLayer.
////////////////////////////////////////////////////////////////////////

FGSwitchLayer::FGSwitchLayer (int w, int h, const SGValue * value,
			      FGInstrumentLayer * layer1,
			      FGInstrumentLayer * layer2)
  : FGInstrumentLayer(w, h), _value(value), _layer1(layer1), _layer2(layer2)
{
}

FGSwitchLayer::~FGSwitchLayer ()
{
  delete _layer1;
  delete _layer2;
}

void
FGSwitchLayer::draw ()
{
  transform();
  if (_value->getBoolValue()) {
    _layer1->draw();
  } else {
    _layer2->draw();
  }
}


// end of panel.cxx
